import torch.nn.functional as F
from torch import nn


# class MLP(nn.Module):
#     def __init__(self):
#         super(MLP, self).__init__()
#         self.flatten = nn.Flatten()
#         self.layer1 = nn.Linear(28 * 28, 64)
#         self.layer2 = nn.Linear(64, 128)
#         self.layer3 = nn.Linear(128, 10)
    
#     def forward(self, x):
#         x = self.flatten(x)
#         x = F.relu(self.layer1(x))
#         x = F.relu(self.layer2(x))
#         x = F.log_softmax(self.layer3(x), dim=1)
#         return x


class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()
        # number of hidden nodes in each layer (512)
        hidden_1 = 128
        hidden_2 = 256
        # linear layer (784 -> hidden_1)
        self.fc1 = nn.Linear(28 * 28, hidden_1)
        # linear layer (n_hidden -> hidden_2)
        self.fc2 = nn.Linear(hidden_1, hidden_2)
        # linear layer (n_hidden -> 10)
        self.fc3 = nn.Linear(hidden_2, 10)
        # dropout layer (p=0.2)
        # dropout prevents overfitting of data
        self.dropout = nn.Dropout(0.2)

    def forward(self, x):
        # flatten image input
        x = x.view(-1, 28 * 28)
        # add hidden layer, with relu activation function
        x = F.relu(self.fc1(x))
        # add dropout layer
        x = self.dropout(x)
        # add hidden layer, with relu activation function
        x = F.relu(self.fc2(x))
        # add dropout layer
        x = self.dropout(x)
        # add output layer
        x = self.fc3(x)
        x = F.log_softmax(x, dim=1)
        return x
        # return x


def create_model():
    return MLP(), nn.modules.loss.CrossEntropyLoss()
